unit uMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, FireDAC.Stan.Intf,
  FireDAC.Stan.Option, FireDAC.Stan.Error, FireDAC.UI.Intf, FireDAC.Phys.Intf,
  FireDAC.Stan.Def, FireDAC.Stan.Pool, FireDAC.Stan.Async, FireDAC.Phys,
  FireDAC.Phys.SQLite, FireDAC.Phys.SQLiteDef, FireDAC.Stan.ExprFuncs,
  FireDAC.Phys.SQLiteWrapper.Stat, FireDAC.VCLUI.Wait, FireDAC.Stan.Param,
  FireDAC.DatS, FireDAC.DApt.Intf, FireDAC.DApt, Data.DB, FireDAC.Comp.DataSet,
  FireDAC.Comp.Client;

type
  TfrmMain = class(TForm)
    Button1: TButton;
    Button2: TButton;
    FDConnection1: TFDConnection;
    FDQuery1: TFDQuery;
    Button3: TButton;
    Memo1: TMemo;
    Button4: TButton;
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure Button4Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  frmMain: TfrmMain;

implementation

{$R *.dfm}

uses
  System.IOUtils, System.JSON.Serializers, System.Rtti,
  PropInspect, INIReaderWriter, DataSetReader;

////////////////////////////////////////////////////////////////////////////////
/// Example 1. Annotating Properties in Property Inspector.
/// Uses JEDI VCL Property Inspector component (https://github.com/project-jedi/jvcl)
type
  TMyObject = class(TPersistent)
  private
    FName: string;
    FValue: string;
  published
    property Name: string read FName write FName;
    [Readable('My value', 'The value of this object')]
    property Value: string read FValue write FValue;
  end;

procedure TfrmMain.Button1Click(Sender: TObject);
var
  LObj: TMyObject;

  procedure DumpObject(AObj: TMyObject);
  begin
    Memo1.Lines.Clear;
    Memo1.Lines.AddPair('Name', AObj.Name);
    Memo1.Lines.AddPair('Value', AObj.Value);
    Memo1.Lines.Add('');
  end;

begin
  LObj := TMyObject.Create;
  try
    LObj.Name := 'John';
    LObj.Value := '40';
    TfrmPropInspect.Execute(LObj);
    DumpObject(LObj);
  finally
    LObj.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
/// Example 2. Persisting Configuration Object in INI file.
type
  TConfigObject = class(TObject)
  private
    FParam1: string;
    FParam2: Integer;
    FParam3: Double;
  public
    [Ini('main', 'param1'), Default('no value')]
    property Param1: string read FParam1 write FParam1;
    [Ini('main', 'param2'), Default(-1)]
    property Param2: Integer read FParam2 write FParam2;
    [Ini('main', 'param3')]
    property Param3: Double read FParam3 write FParam3;
  end;

procedure TfrmMain.Button2Click(Sender: TObject);
var
  LObj: TConfigObject;

  procedure DumpObject(AObj: TConfigObject);
  begin
    Memo1.Lines.Clear;
    Memo1.Lines.AddPair('Param1', AObj.Param1);
    Memo1.Lines.AddPair('Param2', AObj.Param2.ToString);
    Memo1.Lines.AddPair('Param3', AObj.Param3.ToString);
    Memo1.Lines.Add('');
  end;

begin
  LObj := TConfigObject.Create;
  try
    TObjectIniReaderWriter.Load('..\..\data.ini', LObj);
    DumpObject(LObj);

    LObj.Param2 := 20;

    TObjectIniReaderWriter.Save('..\..\data.ini', LObj);
    Memo1.Lines.Add(TFile.ReadAllText('..\..\data.ini'));
  finally
    LObj.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
/// Example 3. Persisting Records in Database.
type
  TEmployee = record
    [DBField('EmployeeID')]
    ID: Integer;
    LastName: string;
    FirstName: string;
    Title: string;
    TitleOfCourtesy: string;
    BirthDate: TDate;
    HireDate: TDate;
    Address: string;
    City: string;
    Region: string;
    PostalCode: string;
    Country: string;
    [DBField('HomePhone')]
    Phone: string;
    Extension: string;
    [DBField(False)]
    Photo: TBytes;
    Notes: string;
    ReportsTo: Integer;
    PhotoPath: string;
  end;

procedure TfrmMain.Button3Click(Sender: TObject);
var
  LReader: TRecordDataSetReader<TEmployee>;
  LEmployee: TEmployee;

  procedure DumpRecord(const ARec: TEmployee);
  var
    LSer: TJsonSerializer;
  begin
    LSer := TJsonSerializer.Create;
    try
      Memo1.Lines.Add(LSer.Serialize(ARec));
    finally
      LSer.Free;
    end;
  end;

begin
  Memo1.Lines.Clear;
  LReader := TRecordDataSetReader<TEmployee>.Create;
  try
    FDQuery1.Close;
    FDQuery1.Open;
    while not FDQuery1.Eof do
    begin
      LReader.Read(FDQuery1, LEmployee);
      DumpRecord(LEmployee);
      FDQuery1.Next;
    end;
  finally
    LReader.Free;
  end;
end;

{$REGION 'Drafts'}
{$HINTS OFF}
{$WARNINGS OFF}
////////////////////////////////////////////////////////////////////////////////
type
  DataAttribute = class(TCustomAttribute);

  [Readable('My enum', 'Set of values from 1 to 3'), Data]
  TEnum = (one, two, three);

  [Readable('My positive', 'Positive integer numbers'), Dataxxx]
  Positive = type Cardinal;

procedure TfrmMain.Button4Click(Sender: TObject);
var
  LCtx: TRttiContext;
  LAttr: ReadableAttribute;
begin
  LAttr := LCtx.GetType(TypeInfo(TEnum)).GetAttribute<ReadableAttribute>;

  LAttr := LCtx.GetType(TypeInfo(Positive)).GetAttribute<ReadableAttribute>;
end;

procedure TestProc;
begin

end;

type
  BadAttribute = class(TCustomAttribute)
  private
    FProc: TProc;
    FEvent: TNotifyEvent;
    FPtr: Pointer;
  public
    constructor Create(const AProc: TProc); overload;
    constructor Create(const AEvent: TNotifyEvent); overload;
    constructor Create(const APtr: Pointer); overload;
    constructor Create(const AArr: array of Byte); overload;
  end;

//  [Bad([1, 2, 3])]
//  Negative = type Integer;

constructor BadAttribute.Create(const AProc: TProc);
begin

end;

constructor BadAttribute.Create(const AEvent: TNotifyEvent);
begin

end;

constructor BadAttribute.Create(const APtr: Pointer);
begin

end;

constructor BadAttribute.Create(const AArr: array of Byte);
begin

end;

{$ENDREGION}

end.
